/*
 * Copyright (c) 2008,2010-2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Thu Jan 10 2008
 */

#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>
#include <err.h>

#include "viam/viam_acq_driver.h"
#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include <cxtypes.h>
# include <cv.h>
# include <highgui.h>
#endif


/* --- local data ---------------------------------------------------- */

typedef struct viamfile_camera_t {
  unsigned int idmin, idmax;
  double fps;
  int width, height;
  int depth, channels;

  unsigned int id;
  char path[PATH_MAX];
  IplImage *frame;
  struct timeval t_acq;
} *viamfile_camera_t;


static int	viamfile_hardware_list(viam_handle_t vh);
static int	viamfile_hardware_attach(viam_handle_t vh,
			viam_camera_t camera);
static void	viamfile_hardware_detach(viam_handle_t vh,
			viam_camera_t camera);
static int	viamfile_camera_configure(viam_handle_t vh,
			viam_camera_t camera);
static int	viamfile_camera_unconfigure(viam_handle_t vh,
			viam_camera_t camera);
static int	viamfile_datatransmit(viam_handle_t vh, viam_camera_t camera,
			viam_onoff_t on);
static int	viamfile_dataflush(viam_handle_t vh, viam_camera_t camera);
static int	viamfile_capture(viam_handle_t vh, viam_camera_t camera);
static char *	viamfile_capturedata(viam_handle_t vh, viam_camera_t camera,
			size_t *bytes, size_t *rowstride);
static struct timeval
		viamfile_capturetimestamp(viam_handle_t vh,
			viam_camera_t camera);
int		viamfile_capturedone(viam_handle_t vh, viam_camera_t camera);
static int	viamfile_queryhwmodes(viam_handle_t vh,
			viam_camera_t camera, int offset,
			viam_hwmode_t *supported, size_t n);
static int	viamfile_queryhwfeature(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			unsigned int *methods, unsigned int *aumodes);
static int	viamfile_sethwfeaturemode(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			viam_filtermethod_t method,
			viam_filterautomode_t aumode);
static int	viamfile_sethwfeaturevalue(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			const viam_filterargs_obj args);


static int	viamfile_getmode(viam_handle_t vh, const char *path,
			viam_hwmode_t *hwmode);
static int	viamfile_hw_fps(viam_hwfps_t swfps, double *hwfps);


const struct viam_acqdev_obj viam_acqdev_entry = {
  "file",
  NULL,
  NULL,
  viamfile_hardware_list,
  viamfile_hardware_attach,
  viamfile_hardware_detach,
  viamfile_camera_configure,
  viamfile_camera_unconfigure,
  viamfile_datatransmit,
  viamfile_dataflush,
  viamfile_capture,
  viamfile_capturedata,
  viamfile_capturetimestamp,
  viamfile_capturedone,
  viamfile_queryhwmodes,
  viamfile_queryhwfeature,
  viamfile_sethwfeaturemode,
  viamfile_sethwfeaturevalue
};


/* --- viamfile_hardware_list ---------------------------------------- */

/** \brief Print available images on the filesystem
 *
 * This function actually prints nothing as any file could virtually be
 * an image...
 *
 * \param[in] vh	Library handle.
 *
 * \return 0 on success.
 */

static int
viamfile_hardware_list(viam_handle_t vh)
{
  return 0;
}


/* --- viamfile_hardware_attach -------------------------------------- */

/** \brief Allocate camera resources for a file pattern
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle to attach.
 *
 * \return 0 on success, or an error.
 */

static int
viamfile_hardware_attach(viam_handle_t vh, viam_camera_t camera)
{
  DIR *dir;
  char *path, *pattern, *uid1, *uid2;
  char conv[4], end;
  struct dirent *de;
  viamfile_camera_t c;
  unsigned int id, idmin, idmax;
  int s;

  /* parse uid into a filename template */
  s = sscanf(viam_camera_uid(camera), "%*[^%]%%%3[0-9ud]%*[^%]%c", conv, &end);
  if (s != 1) {
    warnx("bad filename specifier `%s'", viam_camera_uid(camera));
    return viam_seterrno(vh, Eviamlib_FILE_IO);
  }

  /* get all files matching pattern */
  uid1 = strdup(viam_camera_uid(camera));
  if (!uid1) return viam_seterrno(vh, Eviamlib_MEMORY);
  uid2 = strdup(viam_camera_uid(camera));
  if (!uid2) return viam_seterrno(vh, Eviamlib_MEMORY);
  path = dirname(uid1);
  pattern = basename(uid2);

  dir = opendir(path);
  if (!dir) {
    warn(NULL);
    free(uid1); free(uid2);
    return viam_seterrno(vh, Eviamlib_FILE_IO);
  }

  idmin = (unsigned int)-1; idmax = 0;
  for(de=readdir(dir); de; de=readdir(dir)) {
    if (sscanf(de->d_name, pattern, &id) != 1) continue;

    if (id < idmin) idmin = id;
    if (id > idmax) idmax = id;
  }

  closedir(dir);
  free(uid1); free(uid2);

  /* bail out if nothing found */
  if (idmin > idmax) {
    warnx("no filename matching `%s'", viam_camera_uid(camera));
    return viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);
  }

  /* register camera */
  c = malloc(sizeof(*c));
  if (!c) {
    warnx("unable to allocate memory for camera `%s'", viam_camera_name(camera));
    return viam_seterrno(vh, Eviamlib_MEMORY);
  }
  c->idmin = idmin;
  c->idmax = idmax;
  c->fps = 0.;
  c->id = idmin;
  c->frame = NULL;

  viam_camera_sethwdata(camera, c);

  warnx("camera `%s', indexes from %u to %u",
	viam_camera_name(camera), idmin, idmax);
  return 0;
}


/* --- viamfile_hardware_detach -------------------------------------- */

/** \brief Cleanup dcfile handle
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 */

static void
viamfile_hardware_detach(viam_handle_t vh, viam_camera_t camera)
{
  viamfile_camera_t hw;

  hw = viam_camera_hwdata(camera);
  if (hw) {
    if (hw->frame) cvReleaseImage(&hw->frame);
    free(hw);
  }
}


/* --- viamfile_camera_configure ------------------------------------- */

/** \brief Setup camera with current configuration parameters.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

static int
viamfile_camera_configure(viam_handle_t vh, viam_camera_t camera)
{
  viamfile_camera_t c;
  viam_hwmode_t hwmode;
  int s;

  c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);

  /* configure hardware mode */
  sprintf(c->path, viam_camera_uid(camera), c->id);
  s = viamfile_getmode(vh, c->path, &hwmode);

  if (!viam_camera_hwmode(vh, camera))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (viam_camera_hwmode(vh, camera)->size != hwmode.size ||
      viam_camera_hwmode(vh, camera)->format != hwmode.format ||
      viam_camera_hwmode(vh, camera)->crop != hwmode.crop)
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  if (viamfile_hw_fps(viam_camera_hwmode(vh, camera)->fps, &c->fps))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  if (viam_hwmode_size(
	vh, &hwmode, &c->width, &c->height, &c->depth, &c->channels))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  return 0;
}


/* --- viamfile_camera_unconfigure ----------------------------------- */

/** \brief Reset bank configuration
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

static int
viamfile_camera_unconfigure(viam_handle_t vh, viam_camera_t camera)
{
  return 0;
}


/* --- viam_datatransmit --------------------------------------------- */

int
viamfile_datatransmit(viam_handle_t vh, viam_camera_t camera, viam_onoff_t on)
{
  return 0;
}


/* --- viam_dataflush ------------------------------------------------ */

static int
viamfile_dataflush(viam_handle_t vh, viam_camera_t camera)
{
  return 0;
}


/* --- viamfile_capture ---------------------------------------------- */

static int
viamfile_capture(viam_handle_t vh, viam_camera_t camera)
{
  viamfile_camera_t c;
  struct timeval pt_acq;
  struct timespec w, r;
  double delay;
  int s;

  c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  /* honor framerate */
  pt_acq = c->t_acq;
  gettimeofday(&c->t_acq, NULL);
  delay = c->t_acq.tv_sec + c->t_acq.tv_usec * 1e-6;
  delay -= pt_acq.tv_sec + pt_acq.tv_usec * 1e-6;
  if (delay > 0. && c->fps > 0. && delay < 1./c->fps) {
    w.tv_sec = 1./c->fps - delay;
    w.tv_nsec = ((1./c->fps - delay) - w.tv_sec) * 1e9;

    while(nanosleep(&w, &r) == -1 && errno == EINTR)
      w = r;

    gettimeofday(&c->t_acq, NULL);
  }

  /* read image */
  sprintf(c->path, viam_camera_uid(camera), c->id);
  c->frame = cvLoadImage(c->path, CV_LOAD_IMAGE_UNCHANGED);
  if (!c->frame) {
    c->id++; /* increment id anyway in case of error */
    return viam_seterrno(vh, Eviamlib_FILE_IO);
  }

  /* resize to actual mode size if needed */
  if (c->frame->width != c->width || c->frame->height != c->height) {
    union viam_filterargs_obj args = {{
	.left = 0, .top = 0,
	.right = c->frame->width-1, .bottom = c->frame->height-1,
	.towidth = c->width, .toheight = c->height,
      }};
    IplImage *scaled = cvCreateImage(
      cvSize(c->width, c->height), c->depth, c->channels);
    if (!scaled) {
      cvReleaseImage(&c->frame);
      return viam_seterrno(vh, Eviamlib_MEMORY);
    }

    /* beware: we call geo_scale with a NULL image parameter. That's
     * ok because the current implementation makes no use of this
     * parameter. */
    s = viam_geo_scale(vh, NULL, c->frame, scaled, &args, NULL);

    cvReleaseImage(&c->frame);
    if (s) return s;
    c->frame = scaled;
  }

  return 0;
}


/* --- viamfile_capturedata ------------------------------------------ */

static char *
viamfile_capturedata(viam_handle_t vh, viam_camera_t camera,
		     size_t *bytes, size_t *rowstride)
{
  viamfile_camera_t c = viam_camera_hwdata(camera);
  if (bytes) *bytes = c->frame->imageSize;
  if (rowstride) *rowstride = c->frame->widthStep;
  return (char *)c->frame->imageData;
}


/* --- viamfile_capturetimestamp ------------------------------------- */

static struct timeval
viamfile_capturetimestamp(viam_handle_t vh, viam_camera_t camera)
{
  viamfile_camera_t c = viam_camera_hwdata(camera);
  return c->t_acq;
}


/* --- viamfile_capturedone ------------------------------------------ */

int
viamfile_capturedone(viam_handle_t vh, viam_camera_t camera)
{
  viamfile_camera_t c = viam_camera_hwdata(camera);

  /* increment image index */
  c->id++;

  if (c->frame) {
    cvReleaseImage(&c->frame);
    c->frame = NULL;
  }

  return 0;
}


/* --- viamfile_queryhwmodes ----------------------------------------- */

static int
viamfile_queryhwmodes(viam_handle_t vh, viam_camera_t camera, int offset,
		      viam_hwmode_t *supported, size_t n)
{
  viamfile_camera_t c = viam_camera_hwdata(camera);
  viam_hwmode_t hwmode;
  int i, s, count;

  if (!c) { viam_seterrno(vh, Eviamlib_NOT_CONFIGURED); return 0; }

  /* open current image */
  sprintf(c->path, viam_camera_uid(camera), c->id);
  s = viamfile_getmode(vh, c->path, &hwmode);
  if (s) return s;

  /* generate modes list */
  count = 0;
  for(i=1; i<viam_hwfps_vals && (count-offset)<(signed)n; i++) {
    if (count >= offset) {
      supported[count-offset] = hwmode;
      supported[count-offset].fps = i;
    }
    count++;
  }

  return (count>=offset)?(count-offset>n?n:count-offset):0;
}


/* --- viamfile_queryhwfeature --------------------------------------- */

static int
viamfile_queryhwfeature(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			unsigned int *methods, unsigned int *aumodes)
{
  if (methods) *methods = 0;
  if (aumodes) *aumodes = 0;
  return 0;
}


/* --- viamfile_sethwfeaturemode ------------------------------------- */

/** \brief Set hardware feature mode for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] method	Hardware mode.
 * \param[in] aumode	Auto mode.
 *
 * \return 0 on success.
 */

static int
viamfile_sethwfeaturemode(viam_handle_t vh, viam_camera_t camera,
			  viam_filtertype_t filter,
			  viam_filtermethod_t method,
			  viam_filterautomode_t aumode)
{
  switch(method) {
    case VIAM_FILTER_SOFTWARE:
    case VIAM_FILTER_OFF:
      return 0;

    default: return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  }
}


/* --- viamfile_sethwfeaturevalue ------------------------------------ */

/** \brief Set hardware feature value for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] method	Hardware mode.
 * \param[in] aumode	Auto mode.
 * \param[in] args	Feature value (NULL for AUTO or ONE_PUSH).
 *
 * \return 0 on success.
 */

static int
viamfile_sethwfeaturevalue(viam_handle_t vh, viam_camera_t camera,
			   viam_filtertype_t filter,
			   const viam_filterargs_obj args)
{
  return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
}


/* --- viamfile_getmode ---------------------------------------------- */

static int
viamfile_getmode(viam_handle_t vh, const char *path, viam_hwmode_t *hwmode)
{
  IplImage *image;
  int width, height, depth, channels;
  int i, s;

  /* open current image */
  image = cvLoadImage(path, CV_LOAD_IMAGE_UNCHANGED);
  if (!image) return viam_seterrno(vh, Eviamlib_FILE_IO);

  /* get closest mode */
  hwmode->size = VIAM_HWSZ_INVALID;
  if (image->nChannels > 1)
    hwmode->format = VIAM_HWFMT_RGB888;
  else
    hwmode->format = (image->depth > 8)?VIAM_HWFMT_MONO16:VIAM_HWFMT_MONO8;
  hwmode->crop = VIAM_HW_FIXED;
  hwmode->fps = VIAM_HWFPS_INVALID;
  hwmode->trigger = VIAM_HWTRIGGER_INTERNAL;
  for(i=1; i<viam_hwsize_vals; i++) {
    hwmode->size = i;
    if (viam_hwmode_size(vh, hwmode, &width, &height, &depth, &channels))
      continue;
    if (width > image->width || height > image->height) {
      hwmode->size = i-1;
      break;
    }
  }
  if (hwmode->size == VIAM_HWSZ_INVALID)
    s = viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  else
    s = 0;

  cvReleaseImage(&image);
  return s;
}


/* --- viamfile_fps -------------------------------------------------- */

static const struct {
  viam_hwfps_t sw;
  double hw;
} fpsmap[] = {
  { VIAM_HWFPS_1_875,	1.875},
  { VIAM_HWFPS_3_75,	3.75},
  { VIAM_HWFPS_7_5,	7.5},
  { VIAM_HWFPS_15,	15.},
  { VIAM_HWFPS_30,	30.},
  { VIAM_HWFPS_60,	60.},
  { VIAM_HWFPS_120,	120.},
  { VIAM_HWFPS_240,	240.},
  { VIAM_HWFPS_FREE,	0.}
};

static int
viamfile_hw_fps(viam_hwfps_t swfps, double *hwfps)
{
  int i;

  for(i=0; i<sizeof(fpsmap)/sizeof(fpsmap[0]); i++)
    if (fpsmap[i].sw == swfps) {
      *hwfps = fpsmap[i].hw;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}
