/* $LAAS: bank.c 2008/12/20 16:11:11 tho $ */

/* 
 * Copyright (c) 2007-2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Thu Jan 25 2007
 */

#include <string.h>
#include <stdlib.h>
#include <err.h>

#include "viam_private.h"


/* --- local data ---------------------------------------------------- */

struct viam_bank {
  struct viam_bank *next;

  char *name;			/**<\brief local name */

  int nimages, ncameras;	/**<\brief number of images and cameras */
  viam_image_ref *image;		/**<\brief array of images */
  viam_camera_t *camera;	/**<\brief array of cameras */

  viam_bankcalib_t calibration;	/**<\brief calibration data */

  int configured;		/**<\brief true if configured */
  void *hw_data;		/**<\brief hardware specific data */
  void *usr_data;		/**<\brief user data */
};


/* --- viam_bank_create ---------------------------------------------- */

/** \brief Create an empty group of camera
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Name of the bank.
 *
 * \return bank object or NULL (Eviamlib_MEMORY).
 */

viam_bank_t
viam_bank_create(viam_handle_t vh, const char *name)
{
  viam_bank_t bank;

  bank = malloc(sizeof(*bank));
  if (!bank) {
    warnx("no memory for bank `%s' creation", name);
    viam_seterrno(vh, Eviamlib_MEMORY);
    return NULL;
  }

  bank->name = strdup(name);
  bank->nimages = 0;
  bank->ncameras = 0;
  bank->image = NULL;
  bank->camera = NULL;
  bank->calibration = NULL;
  bank->configured = 0;
  bank->hw_data = NULL;
  bank->usr_data = NULL;

  bank->next = viam_config_banks(vh);
  viam_config_setbanks(vh, bank);

  return bank;
}


/* --- viam_bank_delete ---------------------------------------------- */

/** \brief Destroy a bank
 *
 * \param[in] vh	Library handle
 * \param[in] bank	Bank handle.
 */

void
viam_bank_delete(viam_handle_t vh, viam_bank_t bank)
{
  viam_bank_t b;

  if (bank == viam_config_banks(vh)) {
    viam_config_setbanks(vh, bank->next);
  } else for(b = viam_config_banks(vh); b->next; b = b->next)
    if (bank == b->next) {
      b->next = bank->next;
    }

  free(bank);
}


/* --- viam_bank_cameraadd ------------------------------------------- */

/** \brief Add a new camera to bank configuration
 *
 * \param[in] vh	Library handle
 * \param[in] bank	Bank handle.
 * \param[in] camera	Camera handle.
 * \param[in] name	Name of the resulting image.
 *
 * \return 0 on success
 */

int
viam_bank_cameraadd(viam_handle_t vh, viam_bank_t bank,
		    const viam_camera_t camera, const char *name)
{
  viam_image_ref image;
  void *mem;
  int i;

  /* create space and register new image */
  image = viam_image_create(vh, name, camera);
  if (!image) return viam_errno(vh);

  mem = realloc(bank->image, (bank->nimages+1)*(sizeof(*bank->image)));
  if (!mem) return viam_seterrno(vh, Eviamlib_MEMORY);
  bank->image = mem;
  bank->image[bank->nimages++] = image;

  /* check that camera doesn't belong to the bank */
  for(i=0; i<bank->ncameras; i++) if (bank->camera[i] == camera) {
    /* already here, just return */
    return 0;
  }

  /* create space and register camera */
  mem = realloc(bank->camera, (bank->ncameras+1)*(sizeof(*bank->camera)));
  if (!mem) return viam_seterrno(vh, Eviamlib_MEMORY);
  bank->camera = mem;
  bank->camera[bank->ncameras++] = camera;

  return 0;
}


/* --- viam_bank_getbyname ------------------------------------------- */

/** \brief Get a bank by its local name
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Local name of the bank.
 *
 * \return the bank handle, or NULL if not found
 */

viam_bank_t
viam_bank_getbyname(const viam_handle_t vh, const char *name)
{
  viam_bank_t b;

  for(b = viam_config_banks(vh); b; b = b->next)
    if (!strcmp(b->name, name)) return b;

  viam_seterrno(vh, Eviamlib_NO_SUCH_BANK);
  return NULL;
}


/* --- viam_bank_isconfigured ---------------------------------------- */

/** Return true if bank is configured, false otherwise.
 *
 * \param[in] bank	Bank object.
 *
 * \return true if configured, false otherwise.
 */

int
viam_bank_isconfigured(viam_bank_t bank)
{
  return bank->configured?1:0;
}


/* --- viam_bank mutators and accessors ------------------------------ */

viam_bank_t
viam_bank_next(viam_bank_t bank)
{
  return bank->next;
}

const char *
viam_bank_name(viam_bank_t bank)
{
  return bank->name;
}

int
viam_bank_nimages(viam_bank_t bank)
{
  return bank->nimages;
}

int
viam_bank_ncameras(viam_bank_t bank)
{
  return bank->ncameras;
}

viam_image_ref
viam_bank_image(viam_handle_t vh, viam_bank_t bank, int i)
{
  if (i < 0 || i>= bank->nimages) {
    warnx("trying to access image %d out %d", i, bank->nimages);
    viam_seterrno(vh, Eviamlib_NO_SUCH_IMAGE);
    return NULL;
  }
  return bank->image[i];
}

viam_camera_t
viam_bank_camera(viam_handle_t vh, viam_bank_t bank, int i)
{
  if (i < 0 || i>= bank->ncameras) {
    warnx("trying to access camera %d out %d", i, bank->ncameras);
    viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);
    return NULL;
  }
  return bank->camera[i];
}

viam_bankcalib_t
viam_bank_caldata(viam_bank_t bank)
{
  return bank->calibration;
}

void
viam_bank_setcaldata(viam_bank_t bank, viam_bankcalib_t c)
{
  bank->calibration = c;
}

void
viam_bank_setconfigured(viam_bank_t bank, int configured)
{
  bank->configured = configured;
}

void *
viam_bank_hwdata(viam_bank_t bank)
{
  return bank->hw_data;
}

void
viam_bank_sethwdata(viam_bank_t bank, void *hwdata)
{
  bank->hw_data = hwdata;
}

void *
viam_bank_usrdata(viam_bank_t bank)
{
  return bank->usr_data;
}

void
viam_bank_setusrdata(viam_bank_t bank, void *usrdata)
{
  bank->usr_data = usrdata;
}
