/*
 * Copyright (c) 2007-2008,2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/*
 * Define the few opencv routines that we miss when compiled without
 * opencv support.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include <cxtypes.h>
# include <cv.h>
#endif


/* --- local data ---------------------------------------------------- */

struct viam_image {
  struct viam_image *next;

  char *	name;		/**< local name */
  viam_camera_t camera;		/**< associated camera */
  viam_filter_t flist;		/**< list of filters to apply on image */

  viam_cameracalib_t calibration;/**< calibration data */

  void *	usr_data;	/**<\brief user data */
};


/* --- viam_image_create --------------------------------------------- */

/** \brief Create a new image attached to a given camera
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Image name.
 * \param[in] camera	Camera device from which image gets its data.
 *
 * \return image object
 */

viam_image_ref
viam_image_create(viam_handle_t vh, const char *name, viam_camera_t camera)
{
  viam_image_ref i;

  if (viam_image_getbyname(vh, name)) {
    warnx("image `%s' already exists", name);
    viam_seterrno(vh, Eviamlib_DUPLICATE_ID);
    return NULL;
  }

  i = malloc(sizeof(*i));
  if (!i) {
    warnx("unable to allocate memory for image %s", name);
    viam_seterrno(vh, Eviamlib_MEMORY);
    return NULL;
  }

  i->name = strdup(name);
  i->camera = camera;
  i->flist = NULL;
  i->calibration = NULL;
  i->usr_data = NULL;

  i->next = viam_config_images(vh);
  viam_config_setimages(vh, i);
  return i;
}


/* -- viam_image_geometry -------------------------------------------- */

/** \brief Return image size.
 *
 * The function returns parameters compatible with cvCreateImage()
 * function. The underlying camera must be configured.
 *
 * \param[in] vh	Library handle.
 * \param[in] image	Image object.
 * \param[out] geom	Geometry information.
 *
 * \return 0 if resolution is known and valid, with geom parameter
 * instanciated.
 */

int
viam_image_geometry(viam_handle_t vh, const viam_image_ref image,
		    viam_geometry_t *geom)
{
  viam_filter_t f;
  IplImage *raw;

  if (!image || !viam_camera_setup(viam_image_camera(image)))
    return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);
  if (!geom) return viam_seterrno(vh, Eviamlib_INVALID_VALUE);

  f = viam_image_flist(image);
  if (f) {
    viam_filter_geo(vh, f, geom);
  } else {
    raw = viam_camera_raw(viam_image_camera(image));
    geom->width = raw->width;
    geom->height = raw->height;
    geom->depth = raw->depth;
    geom->channels = raw->nChannels;
    geom->scalex = geom->scaley = 1.;
    geom->offx = geom->offy = 0.;
  }
  return 0;
}


/* --- viam_image_getbyname ------------------------------------------ */

/** \brief Get an image by its local name
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Local name of the image.
 *
 * \return the image handle, or NULL if not found
 */

viam_image_ref
viam_image_getbyname(const viam_handle_t vh, const char *name)
{
  viam_image_ref i;

  for(i = viam_config_images(vh); i; i = i->next)
    if (!strcmp(i->name, name)) return i;

  viam_seterrno(vh, Eviamlib_NO_SUCH_IMAGE);
  return NULL;
}


/* --- viam_image mutators and accessors ----------------------------- */

viam_image_ref
viam_image_next(viam_image_ref image)
{
  return image->next;
}

const char *
viam_image_name(viam_image_ref image)
{
  return image->name;
}

viam_camera_t
viam_image_camera(const viam_image_ref image)
{
  return image?image->camera:NULL;
}

viam_filter_t
viam_image_flist(const viam_image_ref image)
{
  return image->flist;
}

void
viam_image_setflist(viam_image_ref image, viam_filter_t flist)
{
  image->flist = flist;
}

viam_cameracalib_t
viam_image_caldata(const viam_image_ref image)
{
  return image->calibration;
}

void
viam_image_setcaldata(viam_image_ref image, viam_cameracalib_t c)
{
  image->calibration = c;
}


/* ------------------------------------------------------------------- */

/* Follows several opencv replacement functions for when
 * VIAM_OPENCV_SUPPORT is disabled */

#ifndef VIAM_OPENCV_SUPPORT

CvSize
cvSize(int width, int height)
{
  return (CvSize){width, height};
}

IplImage *
cvCreateImage(CvSize size, int depth, int channels)
{
  IplImage *i = cvCreateImageHeader(size, depth, channels);
  cvSetData(i, malloc(i->imageSize), CV_AUTOSTEP);

  if (!i->imageData) cvReleaseImage(&i);
  return i;
}

IplImage *
cvCreateImageHeader(CvSize size, int depth, int channels)
{
  IplImage *i = malloc(sizeof(*i));
  if (!i) return NULL;

  return cvInitImageHeader(i, size, depth, channels, 0, 0);
}

IplImage *
cvInitImageHeader(IplImage *image, CvSize size, int depth, int channels,
		  int origin, int align)
{
  image->nChannels = channels;
  image->depth = depth;
  image->width = size.width;
  image->height = size.height;
  image->widthStep = ((depth & 255) >> 3)*channels * size.width;
  image->imageSize = image->widthStep * image->height;
  image->imageData = NULL;

  return image;
}

void
cvSetData(struct _IplImage *image, void *data, int step)
{
  int line;

  line = ((image->depth & 255) >> 3)*image->nChannels * image->width;
  if (step == CV_AUTOSTEP || step < line)
    image->widthStep = line;
  else
    image->widthStep = step;

  image->imageSize = image->widthStep * image->height;
  image->imageData = data;
}


/* --- cvReleaseImage ------------------------------------------------ */

/* Free image structure */

void
cvReleaseImage(IplImage **image)
{
  if (*image) {
    if ((*image)->imageData) free((*image)->imageData);
    free(*image);
    *image = NULL;
  }
}


/* --- cvSaveImage --------------------------------------------------- */

/* Write an image to filename, in PGM or PPM format */

int
cvSaveImage(const char *filename, const struct _IplImage *image)
{
  FILE *f;
  unsigned char *l, *p;
  int i, j, c;

  f = fopen(filename, "w");
  if (!f) return 1;

  fprintf(f, "P%c %d %d %d\n", image->nChannels==1?'5':'6',
	  image->width, image->height, (1<<image->depth)-1);

  for(i=0, l=image->imageData; i<image->height; i++, l+=image->widthStep) {
    for(j=0, p=l; j<image->width; j++, p+=(image->depth>>3)*image->nChannels)
      for(c=image->nChannels-1; c>=0; c--)
	fwrite(&p[c*(image->depth >> 3)], image->depth >> 3, 1, f);
  }

  fclose(f);
  return 0;
}


/* --- cvLoadImage --------------------------------------------------- */

/* Load an image from filename, in PGM or PPM format */

IplImage *
cvLoadImage(const char *filename, int iscolor)
{
  FILE *f;
  IplImage *image;
  unsigned char *l, *p;
  int i, j, c, s;

  if (iscolor != CV_LOAD_IMAGE_UNCHANGED)
    return NULL;

  f = fopen(filename, "r");
  if (!f) return NULL;

  image = malloc(sizeof(*image));
  if (!image) return NULL;

  s = fscanf(f, "P%1d %d %d %d\n",
	     &image->nChannels, &image->width, &image->height, &image->depth);
  if (s != 4) { free(image); return NULL; }

  if (image->nChannels == 5)
    image->nChannels = 1;
  else
    image->nChannels = 3;

  image->depth++;
  image->depth>>=5;

  image->widthStep = (image->depth >> 3)*image->nChannels * image->width;
  image->imageSize = image->widthStep * image->height;

  image->imageData = malloc(image->imageSize);
  if (!image->imageData) { free(image); return NULL; }

  for(i=0, l=image->imageData; i<image->height; i++, l+=image->widthStep) {
    for(j=0, p=l; j<image->width; j++, p+=(image->depth>>3)*image->nChannels)
      for(c=image->nChannels-1; c>=0; c--)
	fread(&p[c*(image->depth >> 3)], image->depth >> 3, 1, f);
  }

  fclose(f);
  return image;
}

#endif /* VIAM_OPENCV_SUPPORT */
